using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Text;
using System.Windows.Forms;
using Phidgets;
using Phidgets.Events;
using System.Collections;

namespace ActuatorPID
{
    public partial class Form1 : Form
    {
        Graphics g;                     // Graphics object for drawing the graph
        private MotorControl actuator;  // Phidgets MotorControl object for the motor controller board
        
        public Form1()
        {
            InitializeComponent();
        }
        
        // This function plots a point(x,y) on the graph
        public void plotPoint(double x, double y, Color pointColor)
        {
            g.SmoothingMode = SmoothingMode.AntiAlias;
            Pen p = new Pen(pointColor, 1);
            g.DrawEllipse(p, (float)x, 250 - (float)y, 1, 1);
            p.Dispose();
        }

        // This function draws the lines needed to refresh the graph when it loops
        public void wipeGraph(double x)
        {
            Pen p = new Pen(Color.White, 1);
            Pen p2 = new Pen(Color.Gray, 1);
            g.DrawLine(p, (float)x, 0, (float)x, 257);
            g.DrawLine(p2, (float)(x + 0.5), 1, (float)(x + 0.5), 257);
            p.Dispose();
            p2.Dispose();
        }

        // This function clears the graph entirely
        public void clearGraph()
        {
            g = pictureBox1.CreateGraphics();
            g.Clear(Color.LightGray);
        }

        // Detach event handler
        void actuator_Detach(object sender, DetachEventArgs e)
        {
            // Clear motor controller fields
            attachedTxt.Text = "Not Attached";
            nameTxt.Text = "- - -";
            serialTxt.Text = "- - -";
            versionTxt.Text = "- - -";
        }

        // Attach event handler
        void actuator_Attach(object sender, AttachEventArgs e)
        {
            // Populate motor controller fields
            MotorControl attached = (MotorControl)sender;
            attachedTxt.Text = attached.Attached.ToString();
            nameTxt.Text = attached.Name;
            serialTxt.Text = attached.SerialNumber.ToString();
            versionTxt.Text = attached.Version.ToString();
            attached.motors[0].Acceleration = attached.motors[0].AccelerationMax;
            attached.Ratiometric = true;

            // Populate motor selection combo box
            int i;
            for (i = 0; i < attached.motors.Count; i++)
            {
                motorCmb.Items.Add(i);
            }

            // Populate feedback sensor combo box
            for (i = 0; i < attached.sensors.Count; i++)
            {
                analogCmb.Items.Add(i);
            }

            motorCmb.SelectedIndex = 0;
            analogCmb.SelectedIndex = 0;
        }


        //Variables
        double Kp = 1, Ki = 1, Kd = 1; // The three control gains- proportional, integral, and derivative
        double output;                 // output is the result from the control loop calculation
        double input;                  // input is the initial target position chosen by the user
        double feedback;               // feedback is the potentiometer value fed into the control loop
        double error;                  // error is the difference between the target and the actual position
        double errorlast;              // errorlast is the error in the previous iteration of the control loop
        double integral;               // integral is the integral term of the control loop
        double derivative;             // derivative is the derivative term of the control loop
        double dt;                     // dt is the amount of time between iterations of the control loop
        double maxOutput = 100;        // maxOutput is the selected maximum duty cycle of the motor controller
        double deadBand = 0;           // deadBand is the amount of error that can still be present when the motor stops
        bool started = false;          // started controls whether or not the control loop is active
        double graphTimeCount = 0;     // graphTimeCount is the X position variable for the graph

        // This function does the control system calculations and sets output to the duty cycle that the motor needs to run at.
        void PID()
        {
            // Calculate how far we are from the target
            errorlast = error;
            error = input - feedback;

            // If the error is within the specified deadband, and the motor is moving slowly enough,
            // Or if the motor's target is a physical limit and that limit is hit (within deadband margins),
            if ((Math.Abs(error) <= deadBand && Math.Abs(output) < 5)
                || (input < (deadBand + 1 + (double)svMinTxt.Value) && feedback < (deadBand + 1 + (double)svMinTxt.Value))
                || (input > ((double)svMaxTxt.Value - (1 + deadBand)) && feedback > ((double)svMaxTxt.Value - (1 + deadBand))))
            {
                // Stop the motor
                output = 0;
                error = 0;
            }
            else
            {
                // Else, update motor duty cycle with the newest output value
                // This equation is a simple PID control loop
                output = ((Kp * error) + (Ki * integral)) + (Kd * derivative);
                errorTxt.Text = error.ToString();
            }


            //Prevent output value from exceeding maximum output specified by user, otherwise accumulate the integral
            if (output >= maxOutput)
                output = maxOutput;
            else if (output <= -maxOutput)
                output = -maxOutput;
            else
                integral += (error * dt);

            derivative = (error - errorlast) / dt;

            //Update output textbox
            outputTxt.Text = Math.Round(output,2).ToString();

            //Update actual position text box
            feedbackTxt.Text = feedback.ToString();
            
            //Update error text box
            if (targetTxt.Text != "" && feedbackTxt.Text != "")
            {
                errorTxt.Text = Convert.ToString(Convert.ToInt32(targetTxt.Text) - Convert.ToInt32(feedbackTxt.Text));
            }

            // If the time counter of the graph exceeds the width of the frame, loop back to the beginning
            if (graphTimeCount > 800)
                graphTimeCount = 0;

            // Draw the wiper line for the graph
            wipeGraph(graphTimeCount);

            // Plot the points for the position, target, and deadbands
            plotPoint(graphTimeCount, (double)svMinTxt.Value / 4, Color.Black);
            plotPoint(graphTimeCount, (double)svMaxTxt.Value / 4, Color.Black);
            plotPoint(graphTimeCount, feedback / 4, Color.Red);
            plotPoint(graphTimeCount, (input + deadBand) / 4, Color.Yellow);
            plotPoint(graphTimeCount, (input - deadBand) / 4, Color.Yellow);
            plotPoint(graphTimeCount, input / 4, Color.Green);

            // Increment the graph time counter
            graphTimeCount += 0.5;

        }

        // This is the event handler for the attached feedback sensor
        void actuator_SensorUpdate(object sender, SensorUpdateEventArgs e)
        {
            // If the selected sensor input is the one that triggered this handler,
            if ((e.Index == 0 && analogCmb.Text == "0") || (e.Index == 1 && analogCmb.Text == "1"))
            {
                //Store sensor value in global variable
                feedback = e.Value;

                //Run the PID loop
                if (started)
                {
                    PID();
                }

                //Update the motor controller with the new duty cycle
                actuator.motors[Convert.ToInt32(motorCmb.Text)].Velocity = Math.Round(output);
            }
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            // Set up motor object and event handlers
            actuator = new MotorControl();
            actuator.Attach += new AttachEventHandler(actuator_Attach);
            actuator.Detach += new DetachEventHandler(actuator_Detach);
            actuator.SensorUpdate += new SensorUpdateEventHandler(actuator_SensorUpdate);

            // Try to open the motor controller
            // If this fails, the user can retry by reconnecting the USB cable, triggering an attach event
            try
            {
                actuator.open();
                attachedTxt.Text = "Waiting for Attach...";
                actuator.waitForAttachment(10000);
            }
            catch
            {
                attachedTxt.Text = "Attach Timed Out.";
            }

        }

        // This function gets the target position from the slider and starts the PID control loop
        private void startButton_Click(object sender, EventArgs e)
        {
            //Initialize graph and flags
            clearGraph();
            graphTimeCount = 0;
            started = true;

            //Draw Axis label for graph
            Graphics h = pictureBox2.CreateGraphics();
            Pen p = new Pen(Color.Black, 1);
            Font drawFont = new Font("Microsoft Sans Serif", (float)9);
            SolidBrush drawBrush = new SolidBrush(Color.Black);
            StringFormat drawFormat = new StringFormat();
            drawFormat.FormatFlags = StringFormatFlags.DirectionVertical;
            h.DrawString("Position", drawFont, drawBrush, 6, 100, drawFormat);
            h.DrawLine(p, 20, 2, 27, 2);
            h.DrawLine(p, 20, 252, 27, 252);

            //Get the target position and display it in the text box
            targetTxt.Text = PositionTrk.Value.ToString();

            //Update rate determines the time interval
            dt = 0.008;         
            try
            {
                // Get the control gains
                Kp = Convert.ToDouble(kpTxt.Text);
                Ki = Convert.ToDouble(kiTxt.Text);
                Kd = Convert.ToDouble(kdTxt.Text);

                // Get the motor variables
                maxOutput = VelocityTrk.Value;
                deadBand = Convert.ToDouble(deadTxt.Text);
                input = PositionTrk.Value;
            }
            catch
            {
                //Do nothing if any input value is not valid
                return;
            }
        }

        // These functions update the labels when the sliders are moved
        private void VelocityTrk_Scroll(object sender, EventArgs e)
        {
            // Move velocity percentage label
            velocityTrkTxt.Text = (VelocityTrk.Value).ToString() + "%";
            velocityTrkTxt.Location = new Point(Convert.ToInt32(Math.Round(VelocityTrk.Value * 2.39)) + 18, 139);
        }

        private void PositionTrk_Scroll(object sender, EventArgs e)
        {
            // Move position percentage label
            PositionTxt.Text = PositionTrk.Value.ToString();
        }

        private void svMinTxt_ValueChanged(object sender, EventArgs e)
        {
            if (svMinTxt.Value > svMaxTxt.Value)
                svMinTxt.Value = svMaxTxt.Value;

            PositionTrk.Minimum = (int)svMinTxt.Value;
            if (PositionTrk.Value < svMinTxt.Value)
                PositionTrk.Value = (int)svMinTxt.Value;
        }

        private void svMaxTxt_ValueChanged(object sender, EventArgs e)
        {
            if (svMaxTxt.Value < svMinTxt.Value)
                svMaxTxt.Value = svMinTxt.Value;

            PositionTrk.Maximum = (int)svMaxTxt.Value;
            if (PositionTrk.Value > svMaxTxt.Value)
                PositionTrk.Value = (int)svMaxTxt.Value;
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            
            actuator.Attach -= actuator_Attach;
            actuator.Detach -= actuator_Detach;
            actuator.SensorUpdate -= actuator_SensorUpdate;

            started = false;
 
            if (actuator.Attached)
            {
                output = 0;
                actuator.motors[Convert.ToInt32(motorCmb.Text)].Velocity = 0;
                Application.DoEvents();
                actuator.close();
                actuator = null;
            }
            
        }
        //////////////////////////////////////
        // Help functions beyond this point //
        //////////////////////////////////////

        private void targetpositionLabel_HelpRequested(object sender, HelpEventArgs hlpevent)
        {
            System.Windows.Forms.MessageBox.Show("This is the position specified by the user for the actuator to move towards. Position is measured from 0 (completely retracted) to 1000 (fully extended).","Help: Target Position");
        }

        private void actualpositionLabel_HelpRequested(object sender, HelpEventArgs hlpevent)
        {
            System.Windows.Forms.MessageBox.Show("This is the actual position of the actuator according to the feedback potentiometer connected to the analog input. Position is measured from 0 (completely retracted) to 1000 (fully extended).", "Help: Actual Position");
        }

        private void errorLabel_HelpRequested(object sender, HelpEventArgs hlpevent)
        {
            System.Windows.Forms.MessageBox.Show("This is the difference between the target position and the actual position. The goal of the control loop is to reduce absolute value of error to less than the deadband value.", "Help: Error");
        }

        private void velocityLabel_HelpRequested(object sender, HelpEventArgs hlpevent)
        {
            System.Windows.Forms.MessageBox.Show("This is the output of the control loop- in this case, the duty cycle of the motor controller. This value ranges from -100 (Full speed reverse) and +100 (Full speed forward).", "Help: Output Velocity");
        }

        private void motorLabel_HelpRequested(object sender, HelpEventArgs hlpevent)
        {
            System.Windows.Forms.MessageBox.Show("This is the index of the motor being controlled. On a controller with only one motor, this will always be zero.", "Help: Motor Index");
        }

        private void analogLabel_HelpRequested(object sender, HelpEventArgs hlpevent)
        {
            System.Windows.Forms.MessageBox.Show("This is the index of the analog input connected to the feedback potentiometer. Only change events from the selected index will be used.", "Help: Sensor Index");
        }

        private void kpLabel_HelpRequested(object sender, HelpEventArgs hlpevent)
        {
            System.Windows.Forms.MessageBox.Show("This is the proportional gain of the control loop. Increasing Kp while keeping other gains constant should result in the motor reaching the target faster, but may also increase the amount it overshoots the target by.", "Help: Kp");
        }

        private void kiLabel_HelpRequested(object sender, HelpEventArgs hlpevent)
        {
            System.Windows.Forms.MessageBox.Show("This is the integral gain of the control loop. Increasing Ki while keeping other gains constant should result in the motor reaching the target faster and more accurately, but may also increase the amount it overshoots the target by and increase the time it takes to settle on the exact target.", "Help: Ki");
        }

        private void kdLabel_HelpRequested(object sender, HelpEventArgs hlpevent)
        {
            System.Windows.Forms.MessageBox.Show("This is the derivative gain of the control loop. Increasing Kd while keeping other gains constant should result in the motor overshooting its target less, and settling on the exact target quicker. However, increasing it too much could result in unstable or oscillating outputs.", "Help: Kd");
        }

        private void deadBandLabel_HelpRequested(object sender, HelpEventArgs hlpevent)
        {
            System.Windows.Forms.MessageBox.Show("This is the amount of acceptable error on either side of the target. For example, a deadband of 5 and a target position of 450 would cause the motor to stop at any position between 445 and 455 inclusive.", "Help: DeadBand");
        }

        private void attachedLabel_HelpRequested(object sender, HelpEventArgs hlpevent)
        {
            System.Windows.Forms.MessageBox.Show("This shows whether or not a Phidget Motor Controller has attached. If you're having trouble with yout device attaching, ensure that it is the specific product that this example is designed for, and try disconnecting and reconnecting the USB cable.", "Help: Attached");
        }

        private void nameLabel_HelpRequested(object sender, HelpEventArgs hlpevent)
        {
            System.Windows.Forms.MessageBox.Show("This is the name of the attached device being used to control the motor.", "Help: Device Name");
        }

        private void serialLabel_HelpRequested(object sender, HelpEventArgs hlpevent)
        {
            System.Windows.Forms.MessageBox.Show("This is the serial number of the device being used to control the motor.", "Help: Serial Number");
        }

        private void versionLabel_HelpRequested(object sender, HelpEventArgs hlpevent)
        {
            System.Windows.Forms.MessageBox.Show("This is the firmware version of the device being used to control the motor.", "Help: Version");
        }

        private void velocitysliderLabel_HelpRequested(object sender, HelpEventArgs hlpevent)
        {
            System.Windows.Forms.MessageBox.Show("This is the velocity limit of the actuator on a scale of 0 to 100% duty cycle. The control loop will not cause the motor to go faster than this limit.", "Help: Velocity Slider");
        }

        private void targetsliderLabel_HelpRequested(object sender, HelpEventArgs hlpevent)
        {
            System.Windows.Forms.MessageBox.Show("This is the target position of the actuator on a scale of 0 to 100% of the full stroke length. Select a position using this slider and press the start button to begin PID control.", "Help: Target Position Slider");
        }

        private void startButton_HelpRequested(object sender, HelpEventArgs hlpevent)
        {
            System.Windows.Forms.MessageBox.Show("Press this button to start PID control. If the actuator isn't moving, make sure the maximum velocity is set to a nonzero value and make sure the actuator's controller is the device that is attached to this program.", "Help: Start Button");
        }

        private void actuatorImage_HelpRequested(object sender, HelpEventArgs hlpevent)
        {
            System.Windows.Forms.MessageBox.Show("This is a linear actuator. This program will not be of much use to you if you don't have one of these.", "Help: Pictures");
        }

        private void pictureBox1_HelpRequested(object sender, HelpEventArgs hlpevent)
        {
            System.Windows.Forms.MessageBox.Show("This is the graphing area. The graph will show the target position, deadbands, and current position in a scroll-refresh format.", "Help: Graph");
        }

    }


}